// tarpini koda vercia i asma, daro vtbl, etc.
// vtbl.newhandler uzhsipildo VM krovimo metu pagal uzhregistruotus
// handlerius.
// ish visu sorso failu gaunamas vienas vm failas.

/* asm instrukcija:
   opkodas - 2b, count - 2 b; target | vali | valf - 4 b;
   taigi instrukcijos ish 4 arba 8 baitu
   count vienetas == 4 baitai
*/

#include <stdio.h>
#include <cstring>
#include <iostream>

#include "tclass.h"
#include "intcode.h"
#include "utils.h"

using namespace std;

/** VERTIMAS intcode->asm **************************************************/

const long MAXADDRS = 2048; // max. labeliu funkcijoje
const long MAXCODE = 32768 * 12; // max. f-jos kodo dydis

struct TJmpAddr {
    long addr;
    TIntInsn *i;
};

void TFunction::assemble(long b)
{
    static uchar asmbuf[MAXCODE];
    static TJmpAddr jmp[MAXADDRS];
    long njmp = 0;
    // kodinam
    size = 0;
    base = b;
    TIntInsn *i = op;
    while (i) {
        *(short*)(asmbuf + size) = (short)i->op;    size += 2; // opkodas
        *(short*)(asmbuf + size) = (short)i->count; size += 2; // count
        // jei dar reikia
        if (i->op == IOP_PUSHif ||
               i->op == IOP_OFSTHIS || i->op == IOP_OFSFR ||
               i->op == IOP_OFSPTR || i->op == IOP_CALLSUPER) {
            *(long*)(asmbuf + size) = i->vali;
            size += 4;
        }
        // jt arba jmp
        else if (i->target) {
            long ii = 0;
            // gal jau kita puse yra ?
            while (ii < njmp && jmp[ii].i != i->target)
                ii++;
            // nera, dedam i jmp masyva
            if (ii >= njmp) {
                if (i->op != JUMPTARGET) // jmp | jz
                    size += 4; // vieta addr
                else
                    size -= 4; // nereikia op ir cnt
                jmp[njmp].addr = size;
                jmp[njmp].i = i;
                njmp++;
            } else if (i->op == JUMPTARGET)
                *(long*)(asmbuf + jmp[ii].addr - 4) = size + base;
            else {
                *(long*)(asmbuf + size) = jmp[ii].addr + base;
                size += 4; // isirashom addr
            }
        }
        // sekanti insn
        i = i->next;
    }
    as = new uchar[size];
    memcpy( as, asmbuf, size );
}

/** VTBL konstr ************************************************************/

long *TClass::vtbl(long& sz)
{
    // jau padarytas
    if (v_ok) {
        sz = v_sz;
        return v_ptr;
    }
    long t;
    long *tp;
    if (parent)
        tp = parent->vtbl(t); // darom tevo vtbl
    else {
        t = 0;     // tevo nera
        tp = NULL;
    }
    v_sz = t + funcs->Size(); // dar savo pridesim
    v_ptr = new long[v_sz];
    if (tp) // kopijuojam tevo
        memcpy(v_ptr, tp, t*sizeof(long));
    TTableIterator it = funcs->items();
    while (it.hasNext()) {
        TFunction *f = (TFunction*)it.next();
        if (f->Builtin() == EB_NONE)
            v_ptr[f->Offset()] = f->Base();
    }
    // nu tai jau kaip jau jo
    v_ok = true;
    sz = v_sz;
    return v_ptr;
}

/** VISKA RASHOM I FAILA ***************************************************/

/* failo struktura:
    long ncode;      // viso kodo dydis baitais
    char code[ncode];    // visas kodas
    long ncl;        // klasiu kiekis
    classinfo cli[ncl];  // kiekvienos klases classinfo
    long nstr;       // stringu masyvo dydis
    char str[nstr];  // stringu masyvas

    struct classinfo {
    long size;    // klases dydis longais
    char name[];      // asciiz klases vardas
    long parent;      // tevines klases id
    long nvtbl;   // vtbl dydis (metodu kiekis)
    long vtbl[nvtbl]; // vtbl (metodu offs visam kode)
    } */

void TClass::save(const CString& n)
{
    long t, i;
    // sukuriam faila
    FILE *f = fopen(n.c_str(), "wb");
    if (!f) {
        errorMsg(strErrCreateFile.c_str(), n.c_str());
        return;
    }
    // kodo dydis - feikas
    t = 0; fwrite(&t, 4, 1, f);
    // begam, suasminam visu klasiu visas f-jas
    for (i = 0; i < TClass::__id; i++) {
        TClass *c = findId(i);
        if (!c) {
            errorMsg(strErrFindClassId.c_str(), i);
            fclose(f);
            return;
        }
        TTableIterator it = c->Funcs()->items();
        while (it.hasNext()) {
            TFunction *fn = (TFunction*)it.next();
            if (fn->Builtin() == EB_NONE) {
                fn->assemble(t);
                t += fn->Size();
                fwrite(fn->As(), fn->Size(), 1, f);
            }
        }
    }
    // irashom tikra kodo dydi
    fseek(f, 0, SEEK_SET);
    fwrite(&t, 4, 1, f);
    fseek(f, 0, SEEK_END);
    // klasiu kiekis
    t = TClass::__id;
    fwrite(&t, 4, 1, f);
    // klasiu info
    for (i = 0; i < TClass::__id; i++) {
        TClass *c = findId(i);
        // objekto dydis
        t = c->Size(); fwrite(&t, 4, 1, f);
        // klases vardas
        fwrite(c->Name().c_str(), strlen(c->Name().c_str())+1, 1, f);
        // parent id
        if (c->Parent())
            t = c->Parent()->Id();
        else
            t = -1;
        fwrite(&t, 4, 1, f);
        // vtbl
        long *ptr = c->vtbl(t);
        fwrite(&t, 4, 1, f);  // nvtbl
        fwrite(ptr, 4, t, f); // vtbl
    }
    // rasom stringu masyvo dydi
    t = TClass::Strings().Size();
    fwrite(&t, 4, 1, f);
    // rasom stringu masyva
    TString *s = TClass::Strings().First();
    while (s) {
        fwrite(s->s.c_str(), s->s.size()+1, 1, f);
        s = s->next;
    }
    // uzhdarom faila
    fclose(f);
}

