// TTable impl.

#include <cstring>
#include <iostream>

#include "collection.h"

/** constr/destr ***********************************************************/

TTable::TTable( long hashSize )
    : mHashSize(1 << hashSize), mFirst(0), mLast(0), mSize(0)
{
    long i;
    mHash = new TBucket[ mHashSize ];
    for (i = 0; i < mHashSize; i++)
        mHash[i].mFirst = mHash[i].mLast = NULL;
    mHashSize--;
}

TTable::~TTable()
{
    delete[] mHash;

    TTableItem *t1, *t2;
    t1 = mFirst;
    while (t1) {
        t2 = t1->mNext;
        delete t1;
        t1 = t2;
    }
}

/** methods ****************************************************************/

void TTable::addItem( TTableItem *item )
{
    // manage list
    mSize++;
    if (!mFirst)
        mFirst = item;
    else {
        mLast->mNext = item;
        item->mPrev = mLast;
    }
    mLast = item;
    // manage hash bucket
    TBucket& b = mHash[ item->getHashCode() & mHashSize ];
    if (!b.mFirst)
        b.mFirst = item;
    else {
        b.mLast->mHashNext = item;
        item->mHashPrev = mLast;
    }
    b.mLast = item;
}

TTableItem *TTable::findItem( const CString& name ) const
{
    // which hash bucket ?
    TBucket& b = mHash[ name.hashCode() & mHashSize ];
    // lsearch in bucket list
    TTableItem *i = b.mFirst;
    while (i && (i->getName() != name))
        i = i->mHashNext;
    return i;
}

