/* FLEX(LEX) failas BiTikZyz skriptu tokenizacijai */

/* ne unixams lexer.cpp aplink #include <unistd.h> reikia
   #ifdef __GNUC__ ir #endif */

/** PRAD. KODAS I FLEX *****************************************************/

%{

class TNode;
class TType;
class TSymbol;
class TFunction;
class TSymTable;
class TArgs;

// inkludai
#include <iostream>

#include "utils/CString.h"
#include "parser.cpp.h"

long lineNumber = 1; // line number skaitiklis

extern "C" int yywrap(); // declared by the caller

bool skipBody;

void getIdentifier();
void getString();
void getNum();
void getOctNum();
void getHexNum();
void getFloNum();
void getChrNum();
void eatCommentCPP();
void eatCommentC();
void eatBody();

// kitaip leakina atminti
void lexerShutDown()
    { yy_delete_buffer(yy_current_buffer); }

%}

/************
 * MAKROSAI *
 ************/

LETTER  [a-zA-Z_]
DIGIT   [0-9]
SIGN    [+-]
IDENT   {LETTER}({LETTER}|{DIGIT})*
STR     \"[^\"]*\"
NUM     "0"|([1-9]{DIGIT}*)
OCTNUM  "0"[0-7]+
HEXNUM  "0"[xX][0-9a-f]+
CHRNUM  \'[^\']*\'
FLONUM  {DIGIT}*"."{DIGIT}*([eE]{SIGN}?{NUM})?
WSPACE  [\0\a\b\f\r\t\v ]+

/*********************
 * LEXERIO TAISYKLES *
 *********************/

%option never-interactive batch 8bit

%%

"bool"   {return LONG;}  /* == long */
"float"  {return FLOAT;}
"long"   |
"int"    {return LONG;}  /* == long */
"void"   {return VOID;}
"string" {return STRING;}

"false"  |
"NULL"   {yylval.num = 0; return NUMBER;}
"PI"     {yylval.flonum = 3.14159265358979323846; return FLONUMBER;}
"true"   {yylval.num = 1; return NUMBER;}

  /*"break"         {STF(); return BREAK;}*/
  /*"case"          {STF(); return CASE;}*/
  /*"continue"      {STF(); return CONTINUE;}*/
  /*"default"       {STF(); return DEFAULT;}*/
"delete"        {return DELETE;}
"do"            {return DO;}
"else"          {return ELSE;}
"for"           {return FOR;}
  /*"goto"          {STF(); return GOTO;}*/
"if"            {return IF;}
"instanceof"    {return INSTANCEOF;}
"native"        {return NATIVE;}
"new"           {return NEW;}
"return"        {return RETURN;}
"sizeof"        {return SIZEOF;}
"static"        {return STATIC;}
"struct"        {return STRUCT;}
"super"         {return SUPER;}
  /*"switch"        {STF(); return SWITCH;}*/
"this"          {return THIS;}
"typedef"       {return TYPEDEF;}
"typeid"        {return TYPEID;}
"until"         {return UNTIL;}
"while"         {return WHILE;}

";"     {return SEMICOLON;}
"."     {return DOT;}
","     {return COMMA;}
":"     {return DDOT;}
"?"     {return QUESTION;}

"["     {return OPEN_BRAC;}
"]"     {return CLOSE_BRAC;}
"("     {return OPEN_PAR;}
")"     {return CLOSE_PAR;}
"{"     {if (skipBody) { eatBody(); return STUFF; } return OPEN_CS;}
"}"     {return CLOSE_CS;}

"+"     {return ADD;}
"-"     {return SUB;}
"*"     {return MUL;}
"/"     {return DIV;}
"%"     {return MOD;}
"<<"    {return SHL;}
">>"    {return SHR;}
"&"     {return AND;}
"|"     {return OR;}
"^"     {return XOR;}
"~"     {return INV;}
"++"    {return INC;}
"--"    {return DEC;}

"="      {return ASSIGN;}
"+="     {return AS_ADD;}
"-="     {return AS_SUB;}
"*="     {return AS_MUL;}
"/="     {return AS_DIV;}
"%="     {return AS_MOD;}
"<<="    {return AS_SHL;}
">>="    {return AS_SHR;}
"&="     {return AS_AND;}
"|="     {return AS_OR;}
"^="     {return AS_XOR;}

"=="    {return B_E;}
"!="    {return B_NE;}
"&&"    {return B_AND;}
"||"    {return B_OR;}
"!"     {return B_NOT;}
"<"     {return B_L;}
">"     {return B_G;}
"<="    {return B_LE;}
">="    {return B_GE;}

{IDENT}   {getIdentifier(); return IDENT;}
{STR}     {getString(); return STRINGVAL;}
{NUM}     {getNum();    return NUMBER;}
{OCTNUM}  {getOctNum(); return NUMBER;}
{HEXNUM}  {getHexNum(); return NUMBER;}
{CHRNUM}  {getChrNum(); return NUMBER;}
{FLONUM}  {getFloNum(); return FLONUMBER;}

"//"    {eatCommentCPP();} /* komentarai */
"/*"    {eatCommentC();}

\n       {lineNumber++;} /* skaiciuojam eilutes */

{WSPACE} {} /* whitespace: nieko */

.        {return ERROR;} /* kitkas: klaida */

%%

/*************
 * ADISHANAS *
 *************/

/* nuima C++ komentarus */
void eatCommentCPP()
{
        long c;
        while ((c = yyinput()) != '\n' && c != EOF) { };
        lineNumber++;
}
/* nuima C komentarus */
void eatCommentC()
{
        long c;
        for (;;) {
            while ((c = yyinput()) != '*' && c != EOF)
                if (c == '\n') lineNumber++;
            if (c == '*') {
                while ((c = yyinput()) == '*') { };
                if (c == '/') break;
            }
        }
}
/* nuima viska tarp { }, ir paskutini { */
void eatBody()
{
    long c;
    do {
        c = yyinput();
        if (c == '{') eatBody();
        else if (c == '/') {
                c = yyinput();
                if (c == '/') eatCommentCPP();
                else if (c == '*') eatCommentC();
        } else if (c == '\n')
                lineNumber++;
    } while (c != '}' && c != EOF);
}

void getIdentifier() { yylval.str = new CString(yytext); }
void getString() { yylval.str = new CString(yytext+1); yylval.str->erase(1); }
void getNum() { yylval.num = strtol(yytext,0,10); }
void getOctNum() { yylval.num = strtol(yytext+1,0,8); }
void getHexNum() { yylval.num = strtol(yytext+2,0,16); }
void getChrNum()
{
        long i, j;
        j = 0;
        for (i = 1; i < yyleng-1; i++) {
                j <<= 8;
                j += (char)(yytext[i]);
        }
        yylval.num = j;
}
void getFloNum() { yylval.flonum = (float)strtod(yytext,0); }

