
// implem:
// TClass, TType, TTypeTable, TSymbol, TSymTable, TFunction, TFunctable

#include <iostream>

#include "tclass.h"
#include "syntree.h"
#include "intcode.h"
#include "utils.h"

using namespace std;

#define BTZID() { for (long ii = 0; ii < i; ii++) f<<"    "; }

extern CString cfgDir;
extern CString cfgParent;
extern char*   cfgVoid;
extern char*   cfgLong;
extern char*   cfgFloat;
extern bool   cfgVerbose;
extern long   staticSize;
extern long   nativeSlots;

/***************************************************************************/
/* TClass */

long TClass::__id = 0;

TStringTable TClass::strings;

// root klase
TClass::TClass(const CString& n)
{
        name = n;
        types = new TTypeTable(true);
        vars  = new TSymTable(true); // su vptr
        funcs = new TFuncTable(true);
        first = last = parent = next = NULL;
        level = 0;
        v_ok = false; v_sz = 0; v_ptr = NULL;
        id = __id++;
        // pridedam "savo" tipa
        type = new TType(name, this);
        types->add(type);
}

TClass::TClass(const CString& n, TClass *p)
{
        name = n;
        types = new TTypeTable;
        vars  = new TSymTable; // vptr turim ish tevo
        funcs = new TFuncTable;
        parent = p;
        first = last = next = NULL;
        level = p->level + 1;
        v_ok = false; v_sz = 0; v_ptr = NULL;
        id = __id++;
        // pridedam "savo" tipa
        type = new TType(name, this);
        types->add(type);
}

TClass::~TClass()
{
        delete types;
        delete vars;
        delete funcs;
        if (v_ok)
                delete[] v_ptr;
        TClass *c = first;
        while (c) {
                TClass *c2 = c->Next();
                delete c;
                c = c2;
        }
}

long TClass::Size() const
{
        long s = vars->Size() - vars->Offset();
        return s;
}

void TClass::add(TClass *c)
{
        if (!first)
                first = c;
        else
                last->Next(c);
        last = c;
}

TClass *TClass::findId(long i) const
{
        TClass *c = (TClass*)this;
        while (c) {
                if (c->Id() == i)
                        return c;
                if (c->first) {
                        TClass *cc = c->first->findId(i);
                        if (cc)
                                return cc;
                }
                c = c->Next();
        }
        return NULL;
}
TClass *TClass::findName(const CString& n) const
{
        TClass *c = (TClass*)this;
        while (c) {
                if (c->Name() == n)
                        return c;
                if (c->first) {
                        TClass *cc = c->first->findName(n);
                        if (cc)
                                return cc;
                }
                c = c->Next();
        }
        return NULL;
}

// visi ieshkojimai - shioje ir hierarchija tevyn klasese

TType *TClass::findType(const CString& n) const
{
        TClass *c = (TClass*)this;
        TType *p = NULL;
        do {
                p = c->Types()->find(n);
                c = c->Parent();
        } while (c && !p);
        return p;
}

TSymbol *TClass::findSym(const CString& n) const
{
        TClass *c = (TClass*)this;
        TSymbol *p = NULL;
        do {
                p = c->Vars()->find(n);
                c = c->Parent();
        } while (c && !p);
        return p;
}

TFunction *TClass::findFunc(const CString& n) const
{
        TClass *c = (TClass*)this;
        TFunction *p = NULL;
        do {
                p = c->Funcs()->find(n);
                c = c->Parent();
        } while (c && !p);
        return p;
}

void TClass::addType(TType *p)
{
    if (findType(p->getName()))
        errorMsg(strErrDefined.c_str(), "tipas", p->getName().c_str());
    else
        types->add(p);
}

void TClass::addSym(TSymbol *p)
{
    if (findSym(p->getName()))
        errorMsg(strErrDefined.c_str(), "kint.", p->getName().c_str());
    else
        vars->add(p);
}

void TClass::addFunc(TFunction *p)
{
    TFunction *of = findFunc(p->getName());
    if (of)
        funcs->replace(p, of);
    else
        funcs->add(p);
}

// ishveda c++ klases (ir visu vaikiniu) forwarda
void TClass::printFwd(ostream& f)
{
    TClass *c = this;
    while (c) {
        f<<"class "<<c->Name()<<"; static const "<<cfgLong<<" ID"<<c->Name()
                   <<" = "<<c->Id()<<';'<<endl;
        if (c->First())
            c->First()->printFwd(f);
        c = c->Next();
    }
}
// ishveda c++ klases (ir visu vaikiniu) definitiona
void TClass::printDef(ostream& f)
{
    TClass *c = this;
    while (c) {
        f<<"// "<<c->Name()<<" (id="<<c->Id()<<", level="
         <<c->Level()<<")"<<endl;
        f<<"class "<<c->Name()<<' ';
        if (c->Parent())
                f<<": public "<<c->Parent()->Name()<<' ';
        else if (cfgParent[0])
                f<<": public "<<cfgParent<<' ';
        f<<'{'<<endl;
        f<<"public:"<<endl;
        long i = 1;
        // no public constructor!
        c->Types()->print(1,f);
        c->Vars ()->print(1,f);
        c->Funcs()->print(1,f);
        // private dummy constructor
        if (!c->Parent()) {
            f<<"private:"<<endl;
            BTZID();
            f<<c->Name()<<"() { };"<<endl;
        }
        f<<"};"<<endl<<endl;
        if (c->First())
            c->First()->printDef(f);
        c = c->Next();
    }
}
// ishveda c++ klases (ir visu vaikiniu) inlainus
void TClass::printInl(ostream& f)
{
    TClass *c = this;
    while (c) {
        c->Funcs()->print(c, 0, f);
        if (c->First())
            c->First()->printInl(f);
        c = c->Next();
    }
}
// ishveda c++ klases (ir visu vaikiniu) static kintam.
void TClass::printCpp(ostream& f)
{
    TClass *c = this;
    while (c) {
        c->Vars()->print(c, cfgDir, 0, f);
        if (c->First())
            c->First()->printCpp(f);
        c = c->Next();
    }
}
// ishveda c++ klases (ir visu vaikiniu) native metodus
void TClass::printNat(ostream& f)
{
    TClass *c = this;
    while (c) {
        c->Funcs()->print(c, cfgDir, 0, f);
        if (c->First())
            c->First()->printNat(f);
        c = c->Next();
    }
}

/***************************************************************************/
/* TString, TStringTable */

TStringTable::TStringTable()
    : first(NULL), last(NULL), size(0) { };

TStringTable::~TStringTable()
{
    TString *t1, *t2;
    t1 = first;
    while (t1) {
        t2 = t1->next;
        delete t1;
        t1 = t2;
    }
}

// prideda nauja stringa
long TStringTable::add(const CString& s)
{
    TString *t = new TString( s );
    if (!first)
        first = t;
    else
        last->next = t;
    last = t;
    long r = size;
    size += s.size() + 1;
    return r;
}

// suranda stringa, grazina offseta jo
long TStringTable::find(const CString& s) const
{
        TString *t = first;
        long r = 0;
        while (t && (t->s != s)) {
                r += t->s.size() + 1;
                t = t->next;
        }
        if (!t)
                return -1;
        return r;
}

/***************************************************************************/
/* TType, TTypeTable */

TType *TTypeTable::typeRefs[EMAX];

// hashtable of 8
TTypeTable::TTypeTable(bool root)
    : TTable( 3 )
{
        if (!root)
            return;
        // add std. types, if root
        CString s;
        TType *t;
        s = "__error__"; t = new TType(s, EERROR); typeRefs[EERROR] = t; add(t);
        s = cfgVoid    ; t = new TType(s, EVOID ); typeRefs[EVOID ] = t; add(t);
        s = cfgLong    ; t = new TType(s, ELONG ); typeRefs[ELONG ] = t; add(t);
        s = cfgFloat   ; t = new TType(s, EFLOAT); typeRefs[EFLOAT] = t; add(t);
        s = "void*"    ; t = new TType(s, EREF  ); typeRefs[EREF  ] = t; add(t);
        s = "char*"    ; t = new TType(s, ESTR  ); typeRefs[ESTR  ] = t; add(t);
        typeRefs[EUSER] = TypeError;
}

// suranda tipa pagal varda
TType *TTypeTable::find(const CString& n) const
{
    TType *t = (TType*)findItem( n );
    if (!t)
        return NULL;
    // resolve typedef
    while (t && t->Ref())
        t = t->Ref();
    return t;
}

void TTypeTable::print(long i, ostream& f)
{
    TTableIterator it = items();
    if (!it.hasNext())
        return;
    BTZID(); f<<"// types"<<endl;
    while (it.hasNext()) {
        TType *t = (TType*)it.next();
        t->print(i,f);
    }
}

// sukuria builtin tipa
TType::TType(const CString& n, ETypes i)
    : TTableItem( n )
{
        members = NULL;
        id = i;
        switch (id) {
        case ELONG:
        case EREF:
        case ESTR:
        case EFLOAT:  size = 1; break;
        default:      size = 0;
        }
        ref     = NULL;
        cla     = NULL;
}
// sukuria structa
TType::TType(const CString& n, TSymTable *m)
    : TTableItem( n )
{
        members = m;
        size = m->Size();
        id = EUSER;
        ref = NULL;
        cla = NULL;
}
// sukuria klases tipa
TType::TType(const CString& n, TClass *c)
    : TTableItem( n + '*' )
{
        members = NULL;
        size = 1;
        id = EUSER;
        ref = NULL;
        cla = c;
}
// sukuria typedefa
TType::TType(const CString& n, TType *t)
    : TTableItem( n )
{
        members = NULL;
        size = t->Size();
        id = EUSER;
        ref = t;
        cla = NULL;
}

void TType::print(long i, ostream& f)
{
        if (ref) {
                BTZID(); f<<"typedef "<<ref->getName()<<' '<<getName()<<';'<<endl;
        } else if (members) {
                BTZID(); f<<"struct "<<getName()<<'{'<<endl;
                members->print(i+1, f);
                BTZID(); f<<"};"<<endl;
        }
}

TSymbol *TType::findSym(const CString& n) const
{
        if (members) return members->find(n);
        if (cla) return cla->findSym(n);
        return NULL;
}

/***************************************************************************/
/* TSymbol, TSymTable */

void TSymbol::print(long i, ostream& f)
{
        BTZID();
        if (stat) f<<"static ";
        f<<type->getName();
        if (stat) f<<'&';
        f<<' '<<getName();
        if (array >= 0) f<<'['<<array<<']';
        f<<';'<<endl;
}
void TSymbol::print(TClass *c, const CString& vm, long i, ostream& f)
{
        if (!stat) return;
        BTZID();
        f<<type->getName()<<"& "<<c->Name()<<"::"<<getName();
        if (array >= 0) f<<'['<<array<<']';
        f<<" = ("<<type->getName()<<"&)"<<vm<<"_static["<<offset<<"];"<<endl;
}

// symtable, hash of 16 buckets
TSymTable::TSymTable(bool vptr)
    : TTable( 4 ), size(0), offset(0)
{
    if (vptr) {
        CString n = "type";
        add(new TSymbol(TypeLong, n));
    }
}

void TSymTable::add(TSymbol *s)
{
    TTable::addItem( s );
    // suziurim jo dydi
    long sz;
    if (s->Array() < 0)
            sz = s->Type()->Size();
    else
            sz = s->Type()->Size() * s->Array();
    // suziurim jo offseta ir jo lenteles dydi
    if (s->Stat()) { // static
            s->Offset(staticSize);
            staticSize += sz;
    } else { // normal
            s->Offset(size - offset);
            size += sz;
    }
}

void TSymTable::print(long i, ostream& f)
{
    TTableIterator it = items();
    if (!it.hasNext())
        return;
    BTZID(); f<<"// data"<<endl;
    while (it.hasNext()) {
        TSymbol *s = (TSymbol*)it.next();
        s->print(i,f);
    }
}
void TSymTable::print(TClass *c, const CString& vm, long i, ostream& f)
{
    TTableIterator it = items();
    while (it.hasNext()) {
        TSymbol *s = (TSymbol*)it.next();
        s->print(c, vm, i, f);
    }
}

long TSymTable::Size(long n) const
{
    TTableIterator it = items();
    long i = 0, sz = 0;
    while (it.hasNext() && i < n) {
        TSymbol *s = (TSymbol*)it.next();
        if (s->Array() < 0)
            sz += s->Type()->Size();
        else
            sz += s->Type()->Size() * s->Array();
        i++;
    }
    return sz;
}

void TSymTable::Offset(long o)
{
    TTableIterator it = items();
    while (it.hasNext()) {
        TSymbol *s = (TSymbol*)it.next();
        s->Offset(s->Offset() - (o - offset));
    }
    offset = o;
}

/***************************************************************************/
/* TFunction, TFuncTable */

// builtin inline f-jos
static ETypes args0[]  = { EERROR };
static ETypes args1i[] = { ELONG, EERROR };
static ETypes args2i[] = { ELONG, ELONG, EERROR };
static ETypes args3i[] = { ELONG, ELONG, ELONG, EERROR };
static ETypes args1f[] = { EFLOAT, EERROR };
static ETypes args2f[] = { EFLOAT, EFLOAT, EERROR };
static ETypes args3f[] = { EFLOAT, EFLOAT, EFLOAT, EERROR };

TBuiltin TFuncTable::builtins[] = {
// maths
#define MIN(x,y)   (((x) < (y)) ? (x) : (y))            // grazina mazesni
#define MAX(x,y)   (((x) > (y)) ? (x) : (y))            // grazina didesni
#define MID(x,y,z) MAX((x), MIN((y), (z)))              // grazina vidurini
#define SGN(u) ((u)>=0 ? 1:(-1))                        // grazina zenkla
#define ABS(u) ((u)>=0 ? (u):(-(u)))                    // grazina moduli
#define SWAP(aa,bb) { (aa)^=(bb); (bb)^=(aa); (aa)^=(bb); } // sukeicia

{ EB_RAND,      ELONG,  "rand",   args1i, true, "return ::rand() % A;" },
{ EB_ABS,       ELONG,  "abs",    args1i, true, "return ::abs(A);" },
{ EB_SGN,       ELONG,  "sgn",    args1i, true, "return (A>=0) ? 1 : -1;" },
{ EB_MIN,       ELONG,  "imin",   args2i, true, "return (A<B)?A:B;" },
{ EB_MAX,       ELONG,  "imax",   args2i, true, "return (A>B)?A:B;" },
{ EB_CLAMP,     ELONG,  "clamp",  args3i, true, "if (A<B) A=B; else if (A>C) A=C; return A;" },
{ EB_FABS,      EFLOAT, "fabs",   args1f, true, "return (float)::fabs(A);" },
{ EB_FSGN,      EFLOAT, "fsgn",   args1f, true, "return (A>=0.0f) ? 1.0f : -1.0f;" },
{ EB_SIN,       EFLOAT, "sin",    args1f, true, "return (float)::sin(A);" },
{ EB_COS,       EFLOAT, "cos",    args1f, true, "return (float)::cos(A);" },
{ EB_TAN,       EFLOAT, "tan",    args1f, true, "return (float)::tan(A);" },
{ EB_ATAN,      EFLOAT, "atan",   args1f, true, "return (float)::atan(A);" },
{ EB_EXP,       EFLOAT, "exp",    args1f, true, "return (float)::exp(A);" },
{ EB_LOGE,      EFLOAT, "loge",   args1f, true, "return (float)::log(A);" },
{ EB_SQRT,      EFLOAT, "sqrt",   args1f, true, "return (float)::sqrt(A);" },
{ EB_SQR,       EFLOAT, "sqr",    args1f, true, "return A * A;" },
{ EB_FRAND,     EFLOAT, "frand",  args0 , true, "return (float)(::rand())/RAND_MAX;" },
{ EB_FMIN,      EFLOAT, "fmin",   args2f, true, "return (A<B)?A:B;" },
{ EB_FMAX,      EFLOAT, "fmax",   args2f, true, "return (A>B)?A:B;" },
{ EB_FCLAMP,    EFLOAT, "fclamp", args3f, true, "if (A<B) A=B; else if (A>C) A=C; return A;" },
{ EB_LINERP,    EFLOAT, "linerp", args3f, true, "return A + C * (B - A);" },
{ EB_COSERP,    EFLOAT, "coserp", args3f, true, "return A + (1 - (float)(::cos(C * (float)MANO_PI))) * 0.5f * (B - A);" },
{ EB_PRINT,     EVOID,  "print",  args1f, true, "printf(\"  %f\\n\", (double)A);" },
// viskas
{ EB_NONE,      EERROR, NULL,     NULL }
};

// table, hash of 32 buckets
TFuncTable::TFuncTable(bool root)
    : TTable( 5 ), size(0), base(0)
{
    if (root) { // pridedam std. f-jas
        TBuiltin *t = builtins;
        while (t->id != EB_NONE)
            add(new TFunction(t++));
    }
}

void TFuncTable::add(TFunction *f)
{
    TTable::addItem( f );
    f->Offset(size + base);
    if (f->Builtin() == EB_NONE)
        size++;
}

void TFuncTable::replace(TFunction *f, TFunction *of)
{
    // tikrinam, ar sutampa
    if (of->Builtin() != EB_NONE) { // gal builtin
        errorMsg(strErrCantOver.c_str());
        return;
    } if (f->Type() != of->Type()) { // graz. tipai
        errorMsg(strErrOverTypes.c_str());
        warnMsg (strNoteNames.c_str(), f->Type()->getName().c_str(),
                                      of->Type()->getName().c_str());
        return;
    } else if (f->Args() != of->Args()) { // arg. kiekis
        errorMsg(strErrOverCount.c_str());
        warnMsg (strNoteCount.c_str(), f->Args(), of->Args());
        return;
    }
    // arg tipai
    long i = 0;
    TTableIterator its  =  f->Tab()->items();
    TTableIterator itos = of->Tab()->items();
    while (i < f->Args() && its.hasNext() && itos.hasNext()) {
        TSymbol  *s = (TSymbol*) its.next();
        TSymbol *os = (TSymbol*)itos.next();
        if (s->Type() != os->Type()) {
            errorMsg(strErrOverArg.c_str(), i);
            warnMsg (strNoteNames.c_str(), s->Type()->getName().c_str(),
                                          os->Type()->getName().c_str());
            return;
        }
        i++;
    }
    // viskas ok
    // ish tikro pridedam i lentele, ale nedidinam dyzhio ir offseto
    TTable::addItem( f );
    f->Offset(of->Offset());
}

void TFuncTable::Base(long b)
{
    TTableIterator it = items();
    while (it.hasNext()) {
        TFunction *s = (TFunction*)it.next();
        s->Offset(s->Offset() + (b - base));
    }
    base = b;
}

void TFuncTable::print(long i, ostream& f)
{
    TTableIterator it = items();
    if (!it.hasNext()) return;
    BTZID(); f<<"// funcs"<<endl;
    while (it.hasNext()) {
        TFunction *s = (TFunction*)it.next();
        s->print(i, f);
    }
}
void TFuncTable::print(TClass *c, long i, ostream& f)
{
    TTableIterator it = items();
    while (it.hasNext()) {
        TFunction *s = (TFunction*)it.next();
        s->print(c, i, f);
    }
}
void TFuncTable::print(TClass *c, const CString& vm, long i, ostream& f)
{
    TTableIterator it = items();
    while (it.hasNext()) {
        TFunction *s = (TFunction*)it.next();
        s->print(c, vm, i, f);
    }
}

void TFuncTable::optimize() const
{
    TTableIterator it = items();
    while (it.hasNext()) {
        TFunction *s = (TFunction*)it.next();
        s->optimize();
    }
}
void TFuncTable::code() const
{
    TTableIterator it = items();
    while (it.hasNext()) {
        TFunction *s = (TFunction*)it.next();
        s->code();
        s->opticode();
    }
}
void TFuncTable::debug(long n) const
{
    TTableIterator it = items();
    while (it.hasNext()) {
        TFunction *s = (TFunction*)it.next();
        cout<<"      "<<s->getName()<<"():"<<endl;
        s->Tree()->debug(n+1);
        cout<<"      asm("<<s->getName()<<"):"<<endl;
        s->Op()->show();
    }
}

void TFunction::print(long i, ostream& f)
{
    long j = 0;
    // prototipas
    TTableIterator it = tab->items();
    it.next(); // skip retval
    BTZID();
    if (noref) f<<"static ";
	if( !type->Cla() && type->Id() == EUSER )
		f<<type->Belongs()->Name()<<"::";
    f<<type->getName()<<' '<<getName()<<'(';
    while (j < args) {
        TSymbol *s = (TSymbol*)it.next();
        f<<s->Type()->getName()<<' '<<s->getName();
        if (j < args-1) f<<", ";
        j++;
    }
    f<<");"<<endl;
}
void TFunction::print(TClass *c, long i, ostream& f)
{
    if (native >= 0) return; // native f-ja
    // prototipas
    long j = 0;
    TTableIterator it = tab->items();
    it.next(); // skip retval
    BTZID(); f<<"inline ";
	if( !type->Cla() && type->Id() == EUSER )
		f<<type->Belongs()->Name()<<"::";
	f<<type->getName()<<' '<<c->Name()<<"::"<<getName()<<'(';
	while (j < args) {
        TSymbol *s = (TSymbol*)it.next();
        f<<s->Type()->getName()<<' '<<s->getName();
        if (j < args-1) f<<", ";
        j++;
    }
    f<<')'<<endl;
    // builtin f-joms
    if (builtin != EB_NONE) {
        BTZID(); f<<"    { "<<text<<" };"<<endl;
        return;
    }
    // kunas
    BTZID(); f<<'{'<<endl;
    i++;
    // argumentus suvarom i VM steka
    // vieta rezultui
    if (type->Size()) {
        BTZID(); f<<cfgDir<<".grow(sizeof("<<type->getName()<<"));"<<endl;
    }
    long sz = 0; // this
    TTableIterator it2 = tab->items();
    it2.next(); // skip retval
    j = 0;
    while (j < args) {
        TSymbol *s = (TSymbol*)it2.next();
        BTZID(); f<<cfgDir<<".push(&"<<s->getName()<<", sizeof("<<s->getName()<<"));"<<endl;
        sz += s->Type()->Size();
        j++;
    }
    // this
    BTZID(); f<<cfgDir<<".pushi((long)this); ";
    sz++;
    // kviechiam (this, vtblindex)
    f<<cfgDir<<".call(type, "<<offset<<"); ";
    // ishimam argumentus
    f<<cfgDir<<".del("<<sz * 4<<");"<<endl;
    // rezultatas
    if (type->Size()) {
            BTZID(); f<<"return *("<<type->getName()<<"*)"<<cfgDir
                      <<".pop("<<type->Size() * 4<<");"<<endl;
    }
    // galas
    i--;
    BTZID(); f<<"};"<<endl;
}
void TFunction::print(TClass *c, const CString& vm, long i, ostream& f)
{
    if (native < 0) return; // ne native f-ja
    // prototipas
    BTZID(); f<<"void "<<cfgDir<<"_native_"<<native<<"(long *fr)"<<endl;
    BTZID(); f<<'{'<<endl;
    i++;
    // aprasom objekta ir retval
    BTZID(); f<<c->Name()<<" *o = ("<<c->Name()<<"*)fr[-2];"<<endl;
    if (type->Size()) {
        CString ss = "__return_val";
        BTZID();
        if (!type->Cla() && type->Id() == EUSER)
            f<<type->Belongs()->Name()<<"::"<<type->getName()<<" *r = ("
             <<type->Belongs()->Name()<<"::"<<type->getName()<<"*)(fr+("
             <<tab->find(ss)->Offset()<<"));"<<endl;
        else
            f<<type->getName()<<" *r = ("
             <<type->getName()<<"*)(fr+("
             <<tab->find(ss)->Offset()<<"));"<<endl;
    }
    // kvieciam jo metoda, perkeldami visus arg. is VM steko i jo arg.
    BTZID();
    if (type->Size()) f<<"*r = ";
    f<<"o->"<<getName()<<"( ";
    long j = 0;
    TTableIterator it = tab->items();
    it.next(); // skip retval
    while (j < args) {
        TSymbol *s = (TSymbol*)it.next();
        f<<"*(";
        if (!s->Type()->Cla() && s->Type()->Id() == EUSER)
            f<<s->Type()->Belongs()->Name()<<"::";
        f<<s->Type()->getName()<<"*)(fr+("<<s->Offset()<<"))";
        if (j < args-1) f<<", ";
        j++;
    }
    f<<");"<<endl;
    i--;
    BTZID(); f<<"};"<<endl;
}

void TFunction::optimize()
{
    if (!tree) {
        tree = new TEmptyNode();
        return;
    }
    tree = tree->optimize();
}

void TFunction::opticode()
{
    TIntInsn *i;
    do {
        i = op;
        op = op->optimize();
    } while (op != i);
}

// builtin funkcijoms
TFunction::TFunction(TBuiltin *t)
    : TTableItem( t->name )
{
    fill();
    type = TTypeTable::ref(t->ret);
    builtin = t->id;
    noref = t->noref;
    native = -1;
    text = t->text;
    // padarom simboliu (argumentu) lentele
    tab = new TSymTable();
    long cnt = 0;
    tab->add(new TSymbol(TTypeTable::ref(t->ret), strNameVal));
    CString n("A");
    while (t->args[cnt] != EERROR) {
        *(char*)n.c_str() = 'A' + cnt;
        tab->add(new TSymbol(TTypeTable::ref(t->args[cnt]), n));
        cnt++;
    }
    if (!noref)
        tab->add(new TSymbol(TypeLong, strNameThis));
    tab->add(new TSymbol(TypeLong, strNameRet));
    tab->Offset(tab->Size());
    // arg kiekis
    args = cnt;
    // medis
    tree = NULL;
}

TFunction::~TFunction()
{
    delete tab;
    if (tree) tree->del();
    TIntInsn *i1, *i2;
    i1 = op;
    while (i1) {
        i2 = i1->next;
        delete i1;
        i1 = i2;
    }
    delete[] as;
}

