
#include <string.h> /* str???(), mem???() */
#include <stdio.h>  /* malloc(), free() */
#include <stdlib.h> /* fopen(), fgets() */
#include <ctype.h>  /* isspace() */
#include "arocfg.h"

#if defined __cplusplus
extern "C"
{
#endif


// viena konfigos eilute
typedef struct TCfgLine {
	char *name;		// vardas (NULL jeigu tuscia arba komentaras)
	char *data;		// reiksme - kaip eilute
	long  save;		// ar seivinti
	struct TCfgLine *next;	// sekantis
} TCfgLine;

// visas konfigas
typedef struct TCfg {
	TCfgLine *head;        // konfig.eiluciu sarashas
	TCfgLine *tail;        // galas sarasho
	char *saveFile;        // ish kur uzkrovem ?
	long dirty;	       // jei nors kas nors pasikeite
} TCfg;

static TCfg cfg; // konfigas

static long cfgInstalled = 0;

/** irasymas ***************************************************************/

// viska sunaikina, viska surasho
void cfgClose()
{
	TCfgLine *pos, *prev;
	if (cfg.saveFile) {
	    if (cfg.dirty) { // irashom
		FILE *f = fopen(cfg.saveFile, "wt");
		if (f) {
		    pos = cfg.head;
		    while (pos) {   // visas eilutes
			if (!pos->save) { pos = pos->next; continue; }
			if (pos->name) {
			    fputs(pos->name, f);
			    if (pos->name[0] != '[')
				fputs(" = ", f);
			}
			if (pos->data)
			    fputs(pos->data, f);
			fputs("\n", f);
			pos = pos->next;
		    }
		    fclose(f);
		}
	    }
	    free(cfg.saveFile);
	}
	// naikinam sarasha
	pos = cfg.head;
	while (pos) {
	    prev = pos;
	    pos = pos->next;
	    if (prev->name)
		    free(prev->name);
	    if (prev->data)
		    free(prev->data);
	    free(prev);
	}
	cfgInstalled = 0;
}

// inicializuoja viska. jei <load> ir neinicializuota, tai uzkrauna
// default.cfg
void cfgInit(long load)
{
	FILE *f;
	if (load && !cfgInstalled) {
		cfgInstalled = 1;
		f = fopen("default.cfg", "rt");
		if (f) {
			fclose(f);
			cfgFile("default.cfg");
		}
	} else
		cfgInstalled = 1;
}

/** krovimas ***************************************************************/

// susitvarko eilute
static void cfgLine(char *data, char *name, char *val)
{
	long i;
	// praleidzhiam whitespace
	while ((*data) && (isspace(*data)))
	    data++;
	// skaitom varda
	while (*data && (!isspace(*data)) && (*data != '=') && (*data != '#') && (*data != ';'))
	    *name++ = *data++;
	*name = 0;
	// praleidzhiam tarpus ir '='
	while (*data && ((isspace(*data)) || (*data == '=')))
	    data++;
	if (data[0] == '\"')
	    data++;
	strcpy(val, data);
	// gale nuimam whitespace
	i = strlen(val) - 1;
	while ((i >= 0) && (isspace(val[i])))
	    val[i--] = 0;
	if (val[i] == '\"')
	    val[i] = 0;
}

void cfgAddFile(char *loadFile, long save);

// uzkrauna konfiga
void cfgInitFile(char *loadFile, char *saveFile)
{
	if (cfgInstalled)
		cfgClose();
	cfgInit(0);
	cfg.head = cfg.tail = NULL;
	cfg.dirty = 0;
	if (cfg.saveFile) {
		cfg.saveFile = (char*)malloc(strlen(saveFile) + 1);
		strcpy(cfg.saveFile, saveFile);
	} else
		cfg.saveFile = NULL;
	cfgAddFile(loadFile, 1);
}

// dakrauna ir ta faila
void cfgAddFile(char *loadFile, long save)
{
	static char line[256], name[256], val[256];
	FILE *f;
	TCfgLine *p;
	if (!loadFile) return;
	f = fopen(loadFile, "rt");
	if (!f) return;
	while (!feof(f)) {
	    fgets(line, 255, f);
	    if (!line[0]) break;
	    cfgLine(line, name, val);
	    p = (TCfgLine*)malloc(sizeof(TCfgLine));
	    if (name[0]) {
		    p->name = (char*)malloc(strlen(name) + 1);
		    strcpy(p->name, name);
	    } else
		    p->name = NULL;
	    p->data = (char*)malloc(strlen(val) + 1);
	    strcpy(p->data, val);
	    p->save = save;
	    p->next = NULL;
	    if (!cfg.tail)
		cfg.head = cfg.tail = p;
	    else {
		cfg.tail->next = p;
		cfg.tail = p;
	    }
	    // gal include?
	    if (p->name && !strcmp(p->name, "INCLUDE"))
		cfgAddFile(p->data, 0);
	}
	fclose(f);
}

// atidaro ir nuskaito, o irasho pabaigoje irgi i ta pati
void cfgFile(char *fileName)
	{ cfgInitFile(fileName, fileName); }

/** parsinimas *************************************************************/

// uztikrina, kad sekcija yra tarp [] skliaustu
void cfgSectionName(char *in, char *out)
{
	if (in) {
		if (in[0] != '[')
			strcpy(out, "[");
		else
			out[0] = 0;
		strcat(out, in);
		if (out[strlen(out) - 1] != ']')
			strcat(out, "]");
	} else
		out[0] = 0;
}

// suranda ka nors konfige
TCfgLine *cfgFindString(char *section, char *name, TCfgLine **prev)
{
	TCfgLine *p;
	long sect = 1; // 1
	char sectname[256];
	cfgSectionName(section, sectname); // apdedam []
	p = cfg.head;  // ieshkom nuo pradziu
	if (prev)
	    *prev = NULL;
	while (p) {
	    if (p->name) {  // kazkas naudingo
		// kita sekcija
		if (p->name[0]=='[' && p->name[strlen(p->name)-1]==']')
		    sect = !strcmp(sectname, p->name);
		// o gal shita sekcija ?
		if (sect || (name[0] == '['))
		    if (!strcmp(p->name, name))
			return p;
	    }
	    if (prev)
		*prev = p;
	    p = p->next;
	}
	return NULL;
}

// nuskaito eilute ish konfigo
char *cfgGetS(char *section, char *name, char *def)
{
	TCfgLine *p;
	cfgInit(1);
	p = cfgFindString(section, name, NULL);
	if (p)
		return (char*)(p->data ? p->data : "");
	else
		return def;
}

long cfgGetL(char *section, char *name, long def)
{
	char *s = cfgGetS(section, name, NULL);
	if ((s) && (*s))
		return strtol(s, NULL, 0);
	return def;
}

float cfgGetF(char *section, char *name, float def)
{
	char *s = cfgGetS(section, name, NULL);
	if ((s) && (*s))
		return (float)atof(s);
	return def;
}

// nuskaito daug tokenu ish konfigo ( kaip komandine eilute )
// ale vieta, i kuria rodo, yra statine, taigi kita syki kvieciant
// cfgGetArgs bus suporolinta
char **cfgGetArgs(char *section, char *name, long *argc)
{
#define MAX_ARGV 32
	static char buf[256];
	static char *argv[MAX_ARGV];
	long pos, ac;
	char *s = cfgGetS(section, name, NULL);
	if (!s) {
		*argc = 0;
		return NULL;
	}
	strcpy(buf, s);
	pos = ac = 0;
	while ((ac < MAX_ARGV) && buf[pos] && (buf[pos] != '#') && (buf[pos] != ';')) {
		while (buf[pos] && (isspace(buf[pos])))
			pos++;
		if (buf[pos] && (buf[pos] != '#') && (buf[pos] != ';')) {
			argv[ac++] = buf + pos;
			while (buf[pos] && (!isspace(buf[pos])))
				pos++;
			if (buf[pos])
				buf[pos++] = 0;
		}
	}
	*argc = ac;
	return argv;
}

/** keitimai ***************************************************************/

// iterpia nauja eilute
TCfgLine *cfgInsertVar(TCfgLine *p, char *name, char *data)
{
	TCfgLine *n = (TCfgLine*)malloc(sizeof(TCfgLine));
	if (name) {
	    n->name = (char*)malloc(strlen(name) + 1);
	    strcpy(n->name, name);
	} else
	    n->name = NULL;
	if (data) {
	    n->data = (char*)malloc(strlen(data) + 1);
	    strcpy(n->data, data);
	} else
	    n->data = NULL;
	if (p) {
	    n->next = p->next;
	    p->next = n;
	    if (cfg.tail == p)
		cfg.tail = n;
	} else {
	    n->next = NULL;
	    cfg.head = n;
	}
	return n;
}

// irasho eilute i konfiga
void cfgSetS(char *section, char *name, char *val)
{
	TCfgLine *p, *prev;
	char sectname[256];
	cfgInit(1);
	p = cfgFindString(section, name, &prev);
	if (p) { // kintamasis yra
		if (val && *val) { // keiciam i nauja reiksme
			if (p->data)
				free(p->data);
			p->data = (char*)malloc(strlen(val) + 1);
			strcpy(p->data, val);
		} else {	// tuscia reiksme - nereikalingas
			if (p->name) free(p->name);
			if (p->data) free(p->data);
			if (prev)
				prev->next = p->next;
			else
				cfg.head = p->next;
			if (cfg.tail == p)
				cfg.tail = prev;
			free(p);
		}
	} else if ((val) && (*val)) { // nieko nera - reikia naujo
		cfgSectionName(section, sectname);
		if (sectname[0]) {
			p = cfgFindString(NULL, sectname, &prev);
			if (!p) {	// raikia naujos sekcijos
				p = cfg.head;
				while (p && p->next)
				    p = p->next;
				if (p && p->data && *p->data)
				    p = cfgInsertVar(p, NULL, NULL);
				p = cfgInsertVar(p, sectname, NULL);
			}
			while (p && p->next && ((p->next->name && *p->next->name) ||
				(p->next->data && *p->next->data)))
				p = p->next;
			p = cfgInsertVar(p, name, val);
		} else {
			p = cfg.head;
			cfgInsertVar(NULL, name, val);
			cfg.head->next = p;
		}
	}
	cfg.dirty = 1; // true
}

void cfgSetL(char *section, char *name, long val)
{
	char buf[64];
	sprintf(buf, "%li", val);
	cfgSetS(section, name, buf);
}

void cfgSetF(char *section, char *name, float val)
{
	char buf[64];
	sprintf(buf, "%f", val);
	cfgSetS(section, name, buf);
}

#if defined __cplusplus
}
#endif

